package com.iden.bms.service;

import cn.hutool.core.bean.BeanUtil;
import cn.hutool.core.collection.CollUtil;

import cn.hutool.core.util.StrUtil;

import com.baomidou.mybatisplus.core.conditions.query.QueryWrapper;
import com.baomidou.mybatisplus.core.conditions.update.UpdateWrapper;
import com.baomidou.mybatisplus.core.metadata.IPage;
import com.baomidou.mybatisplus.extension.plugins.pagination.Page;
import com.iden.common.cache.DictCache;
import com.iden.common.cache.RedisKeyConstant;
import com.iden.common.cache.RedisUtil;
import com.iden.common.entity.*;
import com.iden.common.enums.CredentialsTypeEnum;
import com.iden.common.enums.GenderEnum;
import com.iden.common.enums.PersonTypeEnum;
import com.iden.common.enums.PopulationTypeEnum;
import com.iden.common.exception.BDException;
import com.iden.common.service.*;

import com.iden.common.vo.CrowdVO;
import com.iden.common.vo.PageReqVO;
import com.iden.common.vo.PersonVO;
import com.iden.common.vo.UserLoginedConvertVO;
import org.apache.commons.lang3.StringUtils;
import org.springframework.beans.BeanUtils;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;

import javax.annotation.Resource;
import java.io.File;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;
import java.util.UUID;

/**
 *
 * @author makejava
 * @since 2021-05-21 00:08:38
 */
@Service
public class CrowdService {

    @Resource
    private IdenCrowdService idenCrowdService;
    @Resource
    private IdenPersonCrowdRefService idenPersonCrowdRefService;
    @Resource
    private IdenCommunityService idenCommunityService;
    @Resource
    private IdenPersonService idenPersonService;
    @Resource
    private IdenWarningPersonService idenWarningPersonService;


    @Resource
    private RedisUtil redisUtil;

    @Value("${iden.root:#{null}}")
    private String idenRoot;

    /**
     * 查询区域列表
     * @return
     */
    public List<CrowdVO> lisCrowd4Select(UserLoginedConvertVO loginUser) {

        QueryWrapper<IdenCrowd> queryWrapper = new QueryWrapper<>();
        queryWrapper.lambda().orderByAsc(IdenCrowd::getCode);

        List<IdenCrowd> list = this.idenCrowdService.list(queryWrapper);
        List<CrowdVO> results = new ArrayList<>();
        if(CollUtil.isNotEmpty(list)) {
            list.forEach(item -> {
                CrowdVO resVO = new CrowdVO();
                BeanUtils.copyProperties(item,resVO);
                results.add(resVO);
            });
        }
        return results;
    }


    /**
     * 查询人群列表
     * @return
             */
    public IPage<CrowdVO> listCrowd(String name, UserLoginedConvertVO loginUser, PageReqVO pageReqVo) {
            IPage<IdenCrowd> page = new Page<>(pageReqVo.getCurrent(), pageReqVo.getPageSize());
            QueryWrapper<IdenCrowd> queryWrapper = new QueryWrapper<>();
            queryWrapper.lambda().like(StrUtil.isNotEmpty(name),IdenCrowd::getName,name)
                    .orderByAsc(IdenCrowd::getCode);

            IPage<IdenCrowd> pageRes = this.idenCrowdService.page(page, queryWrapper);
            IPage<CrowdVO> results = new Page<>(pageRes.getCurrent(),pageRes.getSize(),pageRes.getTotal());
            if(CollUtil.isNotEmpty(pageRes.getRecords())){
                List<CrowdVO> list = new ArrayList<>();
                pageRes.getRecords().forEach(item -> {
                    CrowdVO resVO = new CrowdVO();
                    BeanUtils.copyProperties(item,resVO);
                    list.add(resVO);
                });
                results.setRecords(list);
            }
            return results;
    }

    /**
     * 删除人群
     * @param id
     * @return
     */
    public boolean deleteById(Long id){
        QueryWrapper<IdenPerson> queryWrapper = new QueryWrapper<>();
        queryWrapper.apply("  iden_person.id in ( select person_id from iden_person_crowd_ref ipcr where ipcr.crowd_id = " + id + ")");
        int cnt = idenPersonService.count(queryWrapper);
        if(cnt > 0){
            return false;
        }
        return this.idenCrowdService.removeById(id);
    }

    /**
     * 详情
     * @param id
     * @return
     */
    public CrowdVO getCrowdById(Long id){
        IdenCrowd idenCrowd = this.idenCrowdService.getById(id);
        if (idenCrowd!=null){
            CrowdVO resVO = new CrowdVO();
            BeanUtil.copyProperties(idenCrowd,resVO);
            return resVO;
        }
        return null;
    }


    /**
     * 保存人群
     * @param vo
     */
    @Transactional(rollbackFor = Exception.class)
    public int createCrowd(CrowdVO vo, UserLoginedConvertVO loginUser){
        QueryWrapper<IdenCrowd> queryWrapper = new QueryWrapper<>();
        queryWrapper.lambda().eq(IdenCrowd::getName,vo.getName());
        if(idenCrowdService.count(queryWrapper) > 0){
            return 1;
        }
        QueryWrapper<IdenCrowd> queryWrapper2 = new QueryWrapper<>();
        queryWrapper2.lambda().eq(IdenCrowd::getCode,vo.getCode());
        if(idenCrowdService.count(queryWrapper2) > 0){
            return 2;
        }
        List<Long> personIdList = vo.getPersonIdList();
        if(CollUtil.isEmpty(personIdList)){
            return 3;
        }

        //保存人群
        IdenCrowd idenCrowd = new IdenCrowd();
        BeanUtil.copyProperties(vo,idenCrowd);
        idenCrowd.setCreateTime(new Date());

        this.idenCrowdService.save(idenCrowd);

        List<IdenPersonCrowdRef> idenPersonCrowdRefList = new ArrayList<>();
        for(Long personId : personIdList){
            IdenPersonCrowdRef idenPersonCrowdRef = new IdenPersonCrowdRef();
            idenPersonCrowdRef.setCrowdId(idenCrowd.getId());
            idenPersonCrowdRef.setPersonId(personId);
            idenPersonCrowdRefList.add(idenPersonCrowdRef);
        }
        if(CollUtil.isNotEmpty(idenPersonCrowdRefList)){
            //批量保存
            this.idenPersonCrowdRefService.saveBatch(idenPersonCrowdRefList);
        }
        return 0;
    }

    /**
     * 修改人群
     * @param vo
     */
    @Transactional(rollbackFor = Exception.class)
    public void updateCrowd(CrowdVO vo){
        //保存人群
        IdenCrowd idenCrowd = new IdenCrowd();
        BeanUtil.copyProperties(vo,idenCrowd);
        idenCrowd.setModifyTime(new Date());
        this.idenCrowdService.updateById(idenCrowd);

        List<Long> personIdList = vo.getPersonIdList();
        if(CollUtil.isNotEmpty(personIdList)){
            List<IdenPersonCrowdRef> idenPersonCrowdRefList = new ArrayList<>();
            for(Long personId : personIdList){
                IdenPersonCrowdRef idenPersonCrowdRef = new IdenPersonCrowdRef();
                idenPersonCrowdRef.setCrowdId(idenCrowd.getId());
                idenPersonCrowdRef.setPersonId(personId);
                idenPersonCrowdRefList.add(idenPersonCrowdRef);
            }
            if(CollUtil.isNotEmpty(idenPersonCrowdRefList)){
                //批量保存
                this.idenPersonCrowdRefService.saveBatch(idenPersonCrowdRefList);
            }
        }

    }


    /**
     * 查询人员列表
     * @return
     */
    public IPage<PersonVO> listPersonByCrowd(Long crowdId,UserLoginedConvertVO loginUser, PageReqVO pageReqVo) {

        IPage<IdenPerson> page = new Page<>(pageReqVo.getCurrent(), pageReqVo.getPageSize());

        QueryWrapper<IdenPerson> queryWrapper = new QueryWrapper<>();
        queryWrapper.lambda()
                .orderByDesc(IdenPerson::getModifyTime)
                .orderByDesc(IdenPerson::getCreateTime);
        queryWrapper.apply("  iden_person.id in ( select person_id from iden_person_crowd_ref ipcr where ipcr.crowd_id = "+ crowdId + ")");

        IPage<IdenPerson> pageRes = this.idenPersonService.page(page, queryWrapper);
        IPage<PersonVO> results = new Page<>(pageRes.getCurrent(),pageRes.getSize(),pageRes.getTotal());
        if(CollUtil.isNotEmpty(pageRes.getRecords())){
            List<PersonVO> list = new ArrayList<>();
            pageRes.getRecords().forEach(item -> {
                PersonVO resVO = new PersonVO();
                BeanUtils.copyProperties(item,resVO);
                resVO.setPopulationTypeName(PopulationTypeEnum.getValueToName(resVO.getPopulationType()));
                resVO.setTypeName(PersonTypeEnum.getValueToName(resVO.getType()));
                resVO.setGenderName(GenderEnum.getValueToName(resVO.getGender()));
                resVO.setCredentialsTypeName(CredentialsTypeEnum.getValueToName(resVO.getCredentialsType()));
                resVO.setMarriageName(DictCache.getNameByValue("marriage", resVO.getMarriage()));
                resVO.setPolicitalStatusName(DictCache.getNameByValue("policital_status", resVO.getPolicitalStatus()));
                Long communityId1 = resVO.getCommunityId();
                if(communityId1 != null){
                    IdenCommunity idenCommunity = this.idenCommunityService.getById(communityId1);
                    if(idenCommunity != null) {
                        resVO.setCommunityName(idenCommunity.getName());
                    }
                }
                QueryWrapper<IdenPersonCrowdRef> queryWrapper1 = new QueryWrapper<>();
                queryWrapper1.lambda().eq(IdenPersonCrowdRef::getPersonId,resVO.getId());
                List<IdenPersonCrowdRef> listIdenPersonCrowdRef = idenPersonCrowdRefService.list(queryWrapper1);
                StringBuilder sb = new StringBuilder();
                if (CollUtil.isNotEmpty(listIdenPersonCrowdRef)) {
                    for(IdenPersonCrowdRef idenPersonCrowdRef : listIdenPersonCrowdRef){
                        IdenCrowd idenCrowd = idenCrowdService.getById(idenPersonCrowdRef.getCrowdId());
                        if(idenCrowd != null) {
                            sb.append(idenCrowd.getName()).append(",");
                        }
                    }
                }
                String crowdName = sb.toString();
                if(crowdName != null && crowdName.endsWith(",")) {
                    crowdName = crowdName.substring(0, crowdName.length() - 1);
                }
                resVO.setCrowdName(crowdName);
                list.add(resVO);
            });
            results.setRecords(list);
        }
        return results;
    }


    /**
     * 删除人员
     * @param id
     * @return
     */
    @Transactional(rollbackFor = Exception.class)
    public int deletePersonFromCrowd(Long id, Long crowdId) throws BDException {
        int flag = 0;
        String key = RedisKeyConstant.PERSON_DELETE;
        String requestId = UUID.randomUUID().toString();
        boolean result = redisUtil.tryLock(key,requestId,10 * 60);
        try {
            if (result) {

                QueryWrapper<IdenWarningPerson> queryWrapper0 = new QueryWrapper<>();
                queryWrapper0.lambda().eq(IdenWarningPerson::getPersonId,id);
                int cnt = this.idenWarningPersonService.count(queryWrapper0);
                if(cnt > 0){
                    throw new BDException("人员被引用，不能删除");
                }

                IdenPerson idenPerson = this.idenPersonService.getById(id);
                QueryWrapper<IdenPersonCrowdRef> queryWrapper = new QueryWrapper<>();
                queryWrapper.lambda().eq(IdenPersonCrowdRef::getPersonId,idenPerson.getId());
                idenPersonCrowdRefService.remove(queryWrapper);

                if(StringUtils.isEmpty(idenPerson.getType())){ //不属于小区人员和重点人，才能删除person数据
                    String image = idenPerson.getImage();
                    IdenCommunity idenCommunity = this.idenCommunityService.getById(idenPerson.getCommunityId());
                    if (StringUtils.isNotEmpty(image) && idenCommunity != null) {
                        File imageFile = new File(idenRoot + "data/final/person/image/" + idenCommunity.getCode() + "/" + image.substring(image.lastIndexOf("/") + 1));
                        if(imageFile.exists()){
                            imageFile.delete();
                        }
                    }
                    this.idenPersonService.removeById(id);
                    flag = 1;
                }


            } else {
                throw new BDException("不能多人同时删除，稍后再试");
            }
        } catch (Exception e) {
            e.getMessage();

        }  finally {
            redisUtil.releaseLock(key,requestId);
        }
        return flag;
    }

}