package com.iden.bms.service;




import cn.hutool.core.bean.BeanUtil;
import cn.hutool.core.collection.CollUtil;
import com.baomidou.mybatisplus.core.conditions.query.QueryWrapper;
import com.face.monitor.FaceMonitor;
import com.face.monitor.model.FaceModel;
import com.iden.common.facetool.FaceIdenTool;
import com.iden.common.cache.FaceMoniterCache;
import com.iden.common.cache.RedisKeyConstant;
import com.iden.common.cache.RedisUtil;
import com.iden.common.constant.Constants;
import com.iden.common.entity.*;

import com.iden.common.enums.HandleWarningStatusEnum;
import com.iden.common.enums.PersonTypeEnum;
import com.iden.common.facetool.pool.MyFaceMonitor;
import com.iden.common.service.*;

import com.iden.common.util.ByteUtil;
import com.iden.common.util.DateUtils;
import com.iden.common.util.FileUtil;
import com.iden.common.vo.FaceRetrieveResultVO;
import org.apache.commons.lang3.StringUtils;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.stereotype.Service;

import javax.annotation.Resource;
import java.io.File;
import java.math.BigDecimal;
import java.util.*;


/**
 * @Author: stw
 * @Date: 2021/7/14
 * @Desc:
 */
@Service
public class FaceIdenService {
    @Resource
    private IdenFaceImageService idenFaceImageService;
    @Resource
    private IdenCameraService idenCameraService;
    @Resource
    private IdenCommunityService idenCommunityService;
    @Resource
    private IdenPersonService idenPersonService;
    @Resource
    private IdenPersonStrangerService idenPersonStrangerService;

    @Resource
    private RedisUtil redisUtil;

    @Value("${iden.root:#{null}}")
    private String idenRoot;
    @Value("${file.url:#{null}}")
    private String fileUrl;
    @Value("${ffmpeg.path:#{null}}")
    private String ffmpegPath;


    private static final Logger logger = LogManager.getLogger(FaceIdenService.class);

    /**
     * 初始化人脸识别库
     */
    public void initFaceDataset() {
        FaceMonitor faceMonitor = FaceMoniterCache.getFaceMonitor(Constants.FACE_FEAT_PTR_CACHE_KEY);
        if (faceMonitor == null) {
            MyFaceMonitor myFaceMonitor = FaceIdenTool.getMyFaceMonitor();
            faceMonitor = myFaceMonitor.getFaceMonitor();
            FaceMoniterCache.setFaceMonitor(Constants.FACE_FEAT_PTR_CACHE_KEY, faceMonitor);

            List<FaceModel> faceModelList = new ArrayList<>();
            //注册人员
            QueryWrapper<IdenPerson> queryWrapper = new QueryWrapper<>();
            queryWrapper.select("uid", "feat_ptr");
            queryWrapper.lambda().isNotNull(IdenPerson::getFeatPtr);
            List<IdenPerson> idenPersonList = idenPersonService.list(queryWrapper);
            if (CollUtil.isNotEmpty(idenPersonList)) {

                for (int i = 0; i < idenPersonList.size(); i++) {
                    IdenPerson idenPerson = idenPersonList.get(i);
                    FaceModel faceModel = new FaceModel();
                    faceModel.setPersonId(idenPerson.getUid().intValue());
                    faceModel.setName(String.valueOf(idenPerson.getUid()));
                    faceModel.setFeatValue(ByteUtil.hex2Byte(idenPerson.getFeatPtr()));
                    faceModelList.add(faceModel);
                }
            }

            //陌生人
            QueryWrapper<IdenPersonStranger> queryWrapper1 = new QueryWrapper<>();
            queryWrapper1.select("uid", "feat_ptr");
            queryWrapper1.lambda().isNotNull(IdenPersonStranger::getFeatPtr);
            List<IdenPersonStranger> idenPersonStrangerList = this.idenPersonStrangerService.list(queryWrapper1);
            if (CollUtil.isNotEmpty(idenPersonStrangerList)) {

                for (int i = 0; i < idenPersonStrangerList.size(); i++) {
                    IdenPersonStranger idenPersonStranger = idenPersonStrangerList.get(i);
                    FaceModel faceModel = new FaceModel();
                    faceModel.setPersonId(idenPersonStranger.getUid().intValue());
                    faceModel.setName(String.valueOf(idenPersonStranger.getUid()));
                    faceModel.setFeatValue(ByteUtil.hex2Byte(idenPersonStranger.getFeatPtr()));
                    faceModelList.add(faceModel);
                }
            }

            if (CollUtil.isNotEmpty(faceModelList)) {
                FaceModel[] dataset = new FaceModel[faceModelList.size()];
                faceModelList.toArray(dataset);
                faceMonitor.loadFaceDataset(dataset);
            }
        }
    }



    /**
     * 销毁人脸识别库
     */
    public void destroyFaceDataset() {
        FaceIdenTool.releaseEngine(FaceMoniterCache.getFaceMonitor(Constants.FACE_FEAT_PTR_CACHE_KEY));
    }

    /**
     *  处理各摄像头上传的图像，去除没有人脸的、不是正面的、不符合标准的、连拍的图像
     * @throws Exception
     */
    public void handleCameraImage() throws Exception {
        logger.info("FaceIdenService.handleCameraImage start...");

        File originImageDirFile = new File(idenRoot + "data/origin/camera/image");

        //使用摄像头编码做目录名，里面图像的名称上加上拍照时间，
        // 比如  20211222123223_76639ced-6409-11ec-b8f9-fa163e4e1e9f.jpg
        if (originImageDirFile.isDirectory()) {
            File[] cameraCodeDirs = originImageDirFile.listFiles();
            if(cameraCodeDirs != null && cameraCodeDirs.length > 0){
                logger.info("FaceIdenService.handleCameraImage 1...");

                for (File cameraCodedir : cameraCodeDirs) {
                    if (cameraCodedir.isDirectory()) {
                        File[] imgFiles = FileUtil.sortByName(cameraCodedir.listFiles());
                        if (imgFiles != null && imgFiles.length > 0) {

                            String cameraCode = cameraCodedir.getName();
                            QueryWrapper<IdenCamera> queryWrapper = new QueryWrapper<>();
                            queryWrapper.lambda().eq(IdenCamera::getCode, cameraCode);
                            IdenCamera idenCamera = idenCameraService.getOne(queryWrapper);
                            if (idenCamera == null) {
                                continue;
                            }

                            logger.info("FaceIdenService.handleCameraImage 3...");
                            //获取特征码
                            FaceModel[] faceModels = FaceIdenTool.extractFeature(imgFiles);
                            if(faceModels == null || faceModels.length == 0){
                                continue;
                            }
                            logger.info("FaceIdenService.handleCameraImage 4...");
                           // List<IdenFaceImage> idenFaceImageList = new ArrayList<>();
                            for (int i = 0; i < imgFiles.length; i++) {
                                logger.info("FaceIdenService.handleCameraImage 41...");
                                File imgFile = imgFiles[i];
                                String imgFileName = imgFile.getName();
                                IdenFaceImage idenFaceImage = new IdenFaceImage();
                                //识别特征码
                                if (faceModels[i] != null) {
                                    logger.info("FaceIdenService.handleCameraImage 42...");
                                    String featPtr = ByteUtil.byte2Hex(faceModels[i].getFeatValue());
                                    logger.info("FaceIdenService.handleCameraImage 42...featPtr=="+featPtr);
                                    //没有人脸的、不是正面的、不符合标准的,连拍的图片
                                    if (FaceIdenTool.isBad(featPtr) || isContinuation(cameraCode, imgFileName, featPtr)) {
                                        logger.info("FaceIdenService.handleCameraImage 43...");
                                        File discardDir = new File(imgFile.getParentFile().getAbsolutePath().replace("origin", "discard"));
                                        if (!discardDir.exists()) {
                                            discardDir.mkdirs();
                                        }
                                        File finalImgFile = new File(discardDir, imgFileName);
                                        imgFile.renameTo(finalImgFile);//移动到废弃目录
                                        continue;
                                    } else {
                                        logger.info("FaceIdenService.handleCameraImage 44...");
                                        idenFaceImage.setFeatPtr(featPtr);
                                        //存到redis
                                        redisUtil.hset(RedisKeyConstant.HANDLE_CAMERA_IMAGE_FIRST + "_" + cameraCode, imgFileName, featPtr, RedisKeyConstant.HANDLE_CAMERA_IMAGE_FIRST_TIME);
                                    }

                                }
                                logger.info("FaceIdenService.handleCameraImage 5...");
                                idenFaceImage.setCameraId(idenCamera.getId());
                                idenFaceImage.setCommunityId(idenCamera.getCommunityId());
                                String photographPlace = idenCamera.getPlace();
                                if (StringUtils.isEmpty(photographPlace)) {
                                    IdenCommunity idenCommunity = idenCommunityService.getById(idenCamera.getCommunityId());
                                    if (idenCommunity != null) {
                                        photographPlace = idenCommunity.getName() + "(" + idenCamera.getLongitude() + ":" + idenCamera.getLatitude() + ")";
                                    } else {
                                        photographPlace = "(" + idenCamera.getLongitude() + ":" + idenCamera.getLatitude() + ")";
                                    }
                                }
                                idenFaceImage.setPhotographPlace(photographPlace);
                                idenFaceImage.setLongitude(idenCamera.getLongitude());
                                idenFaceImage.setLatitude(idenCamera.getLatitude());

                                String photographTime = imgFileName.substring(0, imgFileName.indexOf("_"));
                                idenFaceImage.setPhotographTime(DateUtils.strToDate(photographTime, "yyyyMMddHHmmss"));
                                idenFaceImage.setDataDate(photographTime.substring(0, 8));

                                File finalDir = new File(imgFile.getParentFile().getAbsolutePath().replace("origin", "final"));
                                if (!finalDir.exists()) {
                                    finalDir.mkdirs();
                                }
                                logger.info("FaceIdenService.handleCameraImage ...finalDir == " + finalDir.getAbsolutePath());
                                logger.info("FaceIdenService.handleCameraImage ...imgFile.getName() == " + imgFileName);
                                File finalImgFile = new File(finalDir, imgFileName);

                                logger.info("FaceIdenService.handleCameraImage ...finalImgFile == " + finalImgFile.getAbsolutePath());

                                imgFile.renameTo(finalImgFile);//移动到最终目录

                                idenFaceImage.setImage(fileUrl + "camera/image/" + cameraCode + "/" + finalImgFile.getName());

                                //和人脸图特征缓存比对特征码，关联personId
                                FaceMonitor faceMonitor = FaceMoniterCache.getFaceMonitor(Constants.FACE_FEAT_PTR_CACHE_KEY);
                                if (faceModels[i] != null && faceMonitor != null) {

                                    FaceRetrieveResultVO faceRetrieveResultVO = FaceIdenTool.getHitResult(faceMonitor, faceModels[i]);
                                    if (faceRetrieveResultVO != null) { //有命中，要么是注册人员，要么是以前的陌生人员
                                        long hitUid = faceRetrieveResultVO.getIndex();
                                        idenFaceImage.setUid(hitUid);

                                        QueryWrapper<IdenPerson> idenPersonQueryWrapper = new QueryWrapper<>();
                                        idenPersonQueryWrapper.lambda().eq(IdenPerson::getUid,hitUid);
                                        IdenPerson idenPerson = idenPersonService.getOne(idenPersonQueryWrapper);
                                        if (idenPerson != null) {
                                            idenFaceImage.setPersonId(idenPerson.getId());
                                            idenFaceImage.setType(idenPerson.getType());
                                            idenFaceImage.setGender(idenPerson.getGender());
                                            float score = faceRetrieveResultVO.getScore();
                                            BigDecimal b = new BigDecimal(score);
                                            double similarity = b.divide(new BigDecimal(100)).setScale(4, BigDecimal.ROUND_HALF_UP).doubleValue();

                                            idenFaceImage.setSimilarity(similarity);
                                        } else {
                                            QueryWrapper<IdenPersonStranger> idenPersonStrangerQueryWrapper = new QueryWrapper<>();
                                            idenPersonStrangerQueryWrapper.lambda().eq(IdenPersonStranger::getUid,hitUid);
                                            IdenPersonStranger idenPersonStranger = idenPersonStrangerService.getOne(idenPersonStrangerQueryWrapper);
                                            if(idenPersonStranger != null) {
                                                idenFaceImage.setPersonId(idenPersonStranger.getId());

                                                idenFaceImage.setType(PersonTypeEnum.STRANGER.getValue());
                                                float score = faceRetrieveResultVO.getScore();
                                                BigDecimal b = new BigDecimal(score);
                                                double similarity = b.divide(new BigDecimal(100)).setScale(4, BigDecimal.ROUND_HALF_UP).doubleValue();
                                                idenFaceImage.setSimilarity(similarity);
                                            }
                                        }
                                    }  else { //没有命中，肯定是新增陌生人
                                        idenFaceImage.setType(PersonTypeEnum.STRANGER.getValue());
                                        idenFaceImage.setUid(redisUtil.generate(RedisKeyConstant.UID_KEY));
                                        //添加陌生人员表数据
                                        IdenPersonStranger idenPersonStranger = new IdenPersonStranger();
                                        BeanUtil.copyProperties(idenFaceImage,idenPersonStranger);
                                        idenPersonStranger.setCreateTime(new Date());
                                        this.idenPersonStrangerService.save(idenPersonStranger);

                                        //追加人脸库缓存
                                        FaceModel[] dataset = new FaceModel[1];
                                        dataset[0] = new FaceModel();
                                        dataset[0].setPersonId(idenFaceImage.getUid().intValue());
                                        dataset[0].setName(String.valueOf(idenFaceImage.getUid()));
                                        dataset[0].setFeatValue(ByteUtil.hex2Byte(idenFaceImage.getFeatPtr()));
                                        faceMonitor.appendFaceDataset(dataset);
                                    }
                                }
                                logger.info("FaceIdenService.handleCameraImage 6...");

                                idenFaceImage.setCreateTime(new Date());
                                idenFaceImageService.save(idenFaceImage);

                            }
                            logger.info("FaceIdenService.handleCameraImage 7...");

                        }
                    }
                }

            }
        }
        logger.info("FaceIdenService.handleCameraImage end");
    }




   //是否连拍
    private boolean isContinuation(String cameraCode,String imgFileName, String featPtr) {
        logger.info("FaceIdenService.isContinuation ...1");
        String photographTime = imgFileName.substring(0, imgFileName.indexOf("_"));
        Long photographTimeLong = DateUtils.strToDate(photographTime,"yyyyMMddHHmmss").getTime();
        //获取data/final/camera/image下图片
        String originImageDir= idenRoot + "data/final/camera/image";
        logger.info("FaceIdenService.isContinuation ...2");
        File[] imgFiles =  FileUtil.getFiles(originImageDir);
        logger.info("FaceIdenService.isContinuation ...2imgFiles==" + imgFiles);
         List<String> featPtrList = new ArrayList<>();
        if (imgFiles != null && imgFiles.length > 0) {
            for(File imgFile : imgFiles) {
                logger.info("FaceIdenService.isContinuation ...21");
                String photographTimeTmp = imgFileName.substring(0, imgFile.getName().indexOf("_"));
                Long photographTimeTmpLong = DateUtils.strToDate(photographTimeTmp,"yyyyMMddHHmmss").getTime();
                logger.info("FaceIdenService.isContinuation ...3");
                //目标图片和以前图片拍照时间相差2分钟内
                if (photographTimeLong <= photographTimeTmpLong + 2 * 60 * 1000) {
                    String featPtrTmp = (String)redisUtil.hget(RedisKeyConstant.HANDLE_CAMERA_IMAGE_FIRST + "_" + cameraCode, imgFile.getName());
                    if(StringUtils.isNotEmpty(featPtrTmp)){
                        featPtrList.add(featPtrTmp);
                    }
                }
            }
            logger.info("FaceIdenService.isContinuation ...4 featPtrList =" + featPtrList);

            if(CollUtil.isNotEmpty(featPtrList)){
                MyFaceMonitor myFaceMonitorTmp = FaceIdenTool.getMyFaceMonitor();
                FaceMonitor faceMonitorTmp = myFaceMonitorTmp.getFaceMonitor();
                FaceModel[] dataset = new FaceModel[featPtrList.size()];
                for(int i = 0; i< featPtrList.size(); i++) {
                    dataset[i] = new FaceModel();
                    dataset[i].setPersonId(i);
                    dataset[i].setName(String.valueOf(i));
                    dataset[i].setFeatValue(ByteUtil.hex2Byte(featPtrList.get(i)));
                }
                faceMonitorTmp.loadFaceDataset(dataset);

                FaceRetrieveResultVO vo = FaceIdenTool.getHitResult(faceMonitorTmp,featPtr);
                logger.info("FaceIdenService.isContinuation ... 5==" + vo);
                FaceIdenTool.returnObjectToPool(myFaceMonitorTmp);
                return vo != null;
            } else {
                return false;
            }

        } else {
            return false;
        }

    }

    public void reloadFaceDataset(){
        //重新加载人脸库缓存
         initFaceDataset();
    }
}
